#ifndef SORTGEN_H_INCLUDED__
#define SORTGEN_H_INCLUDED__

#include "minxp/int64.h"
#include "minxp/array.h"
#include "random.h"

#include <cmath>
#include <cstddef>
#include <functional>
#include <iterator>
#include <vector>
#include <algorithm>


template<class RanIt, class Pred=std::less<typename std::iterator_traits<RanIt>::value_type> >
class killer : public std::binary_function<RanIt,RanIt,bool>
{
	typedef typename std::iterator_traits<RanIt>::value_type T;
	typedef typename std::iterator_traits<RanIt>::difference_type Diff;
public:
	typedef typename std::vector<Diff>::iterator iterator;
	class predicate
	{
	public:
		predicate(killer *k) : host(k) { }
		bool operator()(const Diff& lhs, const Diff& rhs);
	private:
		killer *host;
	};
	friend class predicate;

	killer(RanIt b, RanIt e, Pred p=Pred()) : its(e-b), solid(), first(b), last(e), candidate(e-b), comp(p)
	{
		T gas = T();
		Diff d = 0;
		for( iterator s=its.begin(); b!=e; ++d, ++s, ++b )
			*s = d, *b = gas;
	}
	~killer()
	{
		for( RanIt i=first; i!=last; ++i ) // Find the one that possibly didn't get fixed
			if( *i == T() )
			{
				factory<T>::increment(solid);
				*i = solid;
			}
	}

	inline iterator begin()
		{ return its.begin(); }
	inline iterator end()
		{ return its.end(); }
	inline predicate pred()
		{ return predicate(this); }

private:
	inline void freeze(Diff i)
	{
		factory<T>::increment(solid);
		first[i] = solid;
	}
	inline bool is_gas(Diff i)
		{ return first[i] == T(); }
	inline bool compare(Diff lhs, Diff rhs)
		{ return comp(first[lhs],first[rhs]); }

	typename std::vector<Diff> its;
	typename std::iterator_traits<RanIt>::value_type solid;
	RanIt first, last;
	Diff candidate;
	Pred comp;
};

template<class RanIt, class Pred>
bool killer<RanIt,Pred>::predicate::operator()(const Diff& lhs, const Diff& rhs)
{
	if( host->is_gas(lhs) && host->is_gas(rhs) )
	{
		if( lhs == host->candidate )
		{
			host->freeze(lhs);
			return true;
		}
		else
		{
			host->freeze(rhs);
			return false;
		}
	}
	else
		if( host->is_gas(lhs) )
		{
			host->candidate = lhs;
			return false;
		}
		else if( host->is_gas(rhs) )
		{
			host->candidate = rhs;
			return true;
		}
		else
			return host->compare(lhs,rhs);
}

template<class T, class It>
void mmerge(It b, It e, size_t k)
{
	random_source rs;
	size_t run_size = (b-e)/k;
	size_t rem = (b-e)%k;
	run_size += rem/k;
	size_t big_runs = rem%k;
	if( big_runs )
		++run_size;
	for( It i=b; b!=e; ++b )
	{
		rs >> *b;
		if( static_cast<size_t>(i-b) == run_size )
		{
			std::sort(b,i);
			if( --big_runs == 0 )
				--run_size;
			b = i;
		}
	}
}

template<class T, class It>
void distr(It b, It e, size_t k)
{
	random_source rs;
	It E = b, p = b;
	for( double dd=1.0/k, d=0.0; p!=e; d+=dd, --k )
		for( E+=(e-p)/k; p!=E; ++p )
		{
			double r;
			rs >> r;
			*p = factory<T>::from_scalar(r*dd+d);
		}
}

template<class T, class It>
void disti(It b, It e, size_t k)
{
	random_source rs;
	T *elts = new T[k];
	for( T *p=elts; p!=elts+k; ++p )
		rs >> *p;
	for( ; b!=e; ++b )
	{
		size_t r;
		rs >> r;
		*b = elts[r%k];
	}
}

template<class T, class It>
void fswaps(It b, It e, size_t k)
{
	random_source rs;
	for( It p=b; p!=e; ++p )
		rs >> *p;
	std::sort(b,e);
	for( ; k; --k )
	{
		size_t i, j;
		rs >> i >> j;
		std::iter_swap(b+i%(e-b), b+j%(e-b));
	}
}

template<class T, class RanIt>
void sortgen(RanIt begin, RanIt end, char d)
{
	if( d == 'k' )
	{
		killer<RanIt> k(begin, end);
		std::sort(k.begin(), k.end(), k.pred());
	}
	else
	{
		random_source rs;
		switch( d )
		{
		case 'u':
			for( ; begin!=end; ++begin )
				rs >> *begin;
			break;
		case 'i':
			{
				T r;
				rs >> r;
				for( ; begin!=end; ++begin )
					*begin = r;
			}
			break;
		case 'm':
			mmerge<T>(begin, end, static_cast<size_t>(log(static_cast<float>(end-begin))));
			break;
		case 'M':
			mmerge<T>(begin, end, static_cast<size_t>(sqrt(static_cast<float>(end-begin))));
			break;
		case 'd':
			distr<T>(begin, end, static_cast<size_t>(log(static_cast<float>(end-begin))));
			break;
		case 'D':
			distr<T>(begin, end, static_cast<size_t>(sqrt(static_cast<float>(end-begin))));
			break;
		case 'e':
			disti<T>(begin, end, static_cast<size_t>(log(static_cast<float>(end-begin))));
			break;
		case 'E':
			disti<T>(begin, end, static_cast<size_t>(sqrt(static_cast<float>(end-begin))));
			break;
		case 'r':
			fswaps<T>(begin, end, static_cast<size_t>(log(static_cast<float>(end-begin))));
			break;
		case 'R':
			fswaps<T>(begin, end, static_cast<size_t>(sqrt(static_cast<float>(end-begin))));
			break;
		default:
			throw std::invalid_argument("Unknown distribution");
		}
	}
	//for( uint64_t i=0; i!=size; ++i )
	//	std::cout << A[i] << std::endl;
}

template<class T>
void sortgen(const char *filename, minxp::uint64_t size, char d)
{
	using namespace minxp;
	map_array<T> A(filename, static_cast<size_t>(size), MAP_CREATE | MAP_WRITE /*| MAP_STREAMING*/);
	sortgen<T>(A.begin(),A.end(),d);
}

#endif // __SORTGEN_H_INCLUDED__
