#ifndef ELEMENTS_H_INCLUDED__
#define ELEMENTS_H_INCLUDED__

#include <ostream>
#include <stdexcept>
#include <limits>
#include "minxp/stopwatch.h"

template<class T>
struct factory;

class element
{
	friend struct factory<element>;
	friend std::ostream& operator<<(std::ostream& os, const element& e);
public:
	element() : key(), ptr() { };
	element(unsigned long int key, void* ptr) : key(key), ptr(ptr) { };
	inline bool operator<(const element& rhs) const
		{ return key<rhs.key; }
	inline bool operator==(const element& rhs) const
		{ return key==rhs.key; }

	inline element operator+=(const element& rhs)
		{ key+=rhs.key; return *this; }
private:
	unsigned long int key;
	void* ptr;
};
std::ostream& operator<<(std::ostream& os, const element& e)
{
	return os << '(' << e.key << ',' << e.ptr << ')';
}
namespace std
{
template<>
class numeric_limits<element>
{
public:
	inline static element min()
	{
		return element(std::numeric_limits<unsigned long int>::min(), static_cast<char*>(NULL)+0xF00DF00D);
	}
	inline static element max()
	{
		return element(std::numeric_limits<unsigned long int>::max(), static_cast<char*>(NULL)+0xF00DF00D);
	}
};
} // namespace std

#include <cstdlib> // for strncmp

class record
{
	friend struct factory<record>;
	friend std::ostream& operator<<(std::ostream& os, const record& e);
public:
	inline bool operator<(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) < 0; }
	inline bool operator==(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) == 0; }

	inline record operator+=(const record& rhs)
	{
		for( int i=0; i!=record_size; ++i )
			rec[i] += rhs.rec[i];
		return *this;
	}

	inline char& operator[](int i)
		{ return rec[i]; }
private:
	enum tag_size { record_size = 100 };
	char rec[record_size];
};
std::ostream& operator<<(std::ostream& os, const record& e)
{
	os << '[';
	for( int i=0; i!=record::record_size; ++i )
		os << e.rec[i];
	return os << ']';
}

namespace std
{
template<>
class numeric_limits<record>
{
public:
	inline static record max()
	{
		record ret;
		for( int i=0; i!=100; ++i )
			ret[i] = std::numeric_limits<char>::max();
		return ret;
	}
	inline static record min()
	{
		record ret;
		for( int i=0; i!=100; ++i )
			ret[i] = std::numeric_limits<char>::min();
		return ret;
	}
};
} // namespace std

#include <cmath>

template<class T>
struct factory
{
	inline static T from_scalar(double d) throw( std::invalid_argument )
	{
		if( d < 0.0 || d > 1.0 )
			throw std::invalid_argument("Scalar must be positive and no more than one.");
		return static_cast<T>(d*std::numeric_limits<T>::max()+(1-d)*std::numeric_limits<T>::min());
	}
	inline static void increment(T& t)
		{ ++t; }
};

template<class T>
struct factory<minxp::counting<T> >
{
	inline static minxp::counting<T> from_scalar(double d) throw( std::invalid_argument )
	{
		return static_cast<minxp::counting<T> >(factory<T>::from_scalar(d));
	}
	inline static void increment(minxp::counting<T>& t)
		{ t = static_cast<minxp::counting<T> >(static_cast<T>(t)+1); }
};

template<>
struct factory<element>
{
	inline static element from_scalar(double d) throw( std::invalid_argument )
		{ return element(factory<unsigned long int>::from_scalar(d), reinterpret_cast<char*>(NULL)+0xBAADF00D); }
	inline static void increment(element& t)
		{ ++t.key; }
};

template<>
struct factory<record>
{
	inline static record from_scalar(double d) throw( std::invalid_argument )
	{
		record ret;
		if( d < 0.0 || d > 1.0 )
			throw std::invalid_argument("Scalar must be positive and no more than one.");
		for( int i=0; i!=record::record_size; ++i )
		{
			d = d*std::numeric_limits<char>::max()-d*std::numeric_limits<char>::min();
			ret.rec[i] = static_cast<char>(d+std::numeric_limits<char>::min());
			d -= floor(d);
		}
		return ret;
	}
	inline static void increment(record& t)
	{
		int i = record::record_size-1;
		for( ++t.rec[i]; t.rec[i]==std::numeric_limits<char>::min() && i; --i );
	}
};

#endif // ELEMENTS_H_INCLUDED__
